#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>

static void put_diff(unsigned char *a, unsigned char *b, int size)
{
	int offset = 0;

	while (offset < size) {
		if (a[offset] == b[offset]) {
			offset++;
			continue;
		}

		printf("%08x:", offset);
		while (offset < size && a[offset] != b[offset]) {
			printf("%02x", b[offset]);
			offset++;
		}
		printf("\n");
	}
}


int main(int argc, char **argv)
{
	unsigned char *src, *dst;
	struct stat src_st, dst_st;
	FILE *fp;

	if (argc != 3) {
		fprintf(stderr, "usage: bdiff src dst > bpatchfile\n");
		return 1;
	}

	if (stat(argv[1], &src_st) ||
		stat(argv[2], &dst_st)) {
		perror("stat");
		return 1;
	}

	if (src_st.st_size != dst_st.st_size) {
		fprintf(stderr, "file sizes are different.\n");
		return 1;
	}

	src = malloc(src_st.st_size);
	dst = malloc(dst_st.st_size);

	if (!src || !dst) {
		perror("malloc");
		return 1;
	}

	fp = fopen(argv[1], "r");
	if (fread(src, src_st.st_size, 1, fp) != 1) {
		perror("fread");
		return 1;
	}
	fclose(fp);

	fp = fopen(argv[2], "r");
	if (fread(dst, dst_st.st_size, 1, fp) != 1) {
		perror("fread");
		return 1;
	}
	fclose(fp);

	put_diff(src, dst, src_st.st_size);

	free(src);
	free(dst);

	return 0;
}
