#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>

#define HEX2DEC(c) ((isdigit(c))?(c)-'0':10+(((c)-'A')&0x7))

static int parseLine(unsigned long *addr, unsigned char *outbuf,
					 int outbuf_size,
					 const char *inbuf)
{
	char *p;
	int i;
	int tmp;

	if(!inbuf)
		return -1;

	if(strlen(inbuf) < 11)
		return -1;

	*addr = strtoul(inbuf, &p, 16);
	if(p - inbuf != 8 || *p != ':'){
		printf("invalid patchtext format\n");
		return -1;
	}

	p++;

	i = 0;
	while(p[0] != '\0' && p[1] != '\0'){
		if(!isxdigit(p[0]) || !isxdigit(p[1]))
			break;

		outbuf[i++] = (HEX2DEC(p[0]) << 4) | HEX2DEC(p[1]);
		p+=2;
	}

	if(*p != '\r' && *p != '\n' && *p != '\0')
		return -1;

	return i;
}

void dump(const void *ptr, int size)
{
	int i;
	const unsigned char *p = ptr;

	for(i = 0; i < size; i++){
		if(i % 16 == 0)
			printf("%08x:", i);
		printf(" %02x", *p++);
		if(i % 16 == 15)
			printf("\n");
	}
	if(i % 16)
		printf("\n");
}


int main(int argc, char **argv)
{
	FILE *fp;
	char buf[256];
	unsigned long addr = 0;
	unsigned char patch_buf[256];
	int ret;

	if(argc != 2){
		printf("bpatch [TargetFile] < PatchText\n"
			   "PatchText:\n"
			   "XXXXXXXX:XX[XX...]\n"
			   "\nex:\n"
			   "00001234:ffaabbccddee\n");
		return 1;
	}

	fp = fopen(argv[1], "r+");
	if(!fp){
		perror("open");
		return 1;
	}

	while(fgets(buf, sizeof(buf)-1, stdin) != NULL){
		ret = parseLine(&addr, patch_buf, sizeof(patch_buf), buf);
		if(ret < 0)
			return -1;

		fseek(fp, addr, SEEK_SET);
		fwrite(patch_buf, ret, 1, fp);
	}

	fclose(fp);

	return 0;
}
